#include <fstream>
#include <ShlObj.h>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/xml_parser.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>
#include "configuration.h"
#include "sp_exception.h"
#include "web_content_loader.h"
#include "image_window.h"
#include "url_processor.h"
#include "skype.h"

HINSTANCE configuration::instance_ = 0;
configuration::prop_map configuration::props_;
web_content_loader* configuration::loader_ = 0;
image_window* configuration::displayer_ = 0;

void configuration::set_instance(HINSTANCE instance)
{
	instance_ = instance;
}

HINSTANCE configuration::get_instance()
{
	return instance_;
}

bool configuration::conf_property::save_to_file() const
{
	return save_to_file_;
}

configuration::conf_property::conf_property()
	:save_to_file_(false)
{
}

const configuration::conf_property::value_type& configuration::conf_property::get_variant() const
{
	return prop_;
}

bool configuration::cast_to_bool(const std::string& value)
{
	if(boost::iequals(value, "true"))
		return true;

	if(boost::iequals(value, "false"))
		return false;

	throw sp_exception(sp_exception::internal_error, L"Unknown boolean value");
}

void configuration::clear(bool saved_only)
{
	for(prop_map::const_iterator it = props_.begin(); it != props_.end();)
	{
		if(saved_only && !(*it).second.save_to_file())
			++it;
		else
			props_.erase(it++);
	}
}

void configuration::load(const std::wstring& file_name, const std::wstring& app_name)
{
	std::ifstream stream(prepare_settings_directory(file_name, app_name, false).c_str(), std::ios::in);
	if(!stream.is_open() || stream.bad())
		throw sp_exception(sp_exception::internal_error, L"Unable to open xml file");

	boost::property_tree::ptree pt;
	boost::property_tree::xml_parser::read_xml(stream, pt);

	const boost::property_tree::ptree& settings = pt.get_child("settings");

	for(boost::property_tree::ptree::const_iterator it = settings.begin(); it != settings.end(); ++it)
	{
		if((*it).first != "property")
			continue;

		try
		{
			const boost::property_tree::ptree& prop = (*it).second;
			const std::string value_type(prop.get_child("<xmlattr>.type").get_value<std::string>());
			const std::string name(prop.get_child("<xmlattr>.name").get_value<std::string>());
			if(value_type == "string")
			{
				set_property(name, (*it).second.data(), true);
			}
			else if(value_type == "uint32_t")
			{
				set_property(name, boost::lexical_cast<uint32_t>((*it).second.data()), true);
			}
			else if(value_type == "bool")
			{
				set_property(name, cast_to_bool((*it).second.data()), true);
			}
			else if(value_type == "set")
			{
				string_set val;
				const boost::property_tree::ptree& set_ptree = (*it).second;
				for(boost::property_tree::ptree::const_iterator set_it = set_ptree.begin();
					set_it != set_ptree.end(); ++set_it)
				{
					if((*set_it).first != "elem")
						continue;

					const std::string data((*set_it).second.data());
					if(!data.empty())
						val.insert(data);
				}

				set_property(name, val, true);
			}
			else
			{
				throw sp_exception(sp_exception::internal_error, L"Unknown config data type");
			}
		}
		catch(const std::exception&)
		{
		}
	}
}

struct save_visitor : public boost::static_visitor<>
{
public:
	save_visitor(boost::property_tree::ptree& settings, const std::string& name)
		:settings_(settings),
		name_(name)
	{
	}

	void put_name(boost::property_tree::ptree& prop) const
	{
		prop.put("<xmlattr>.name", name_);
	}

	void put_type(boost::property_tree::ptree& prop, const std::string& value_type) const
	{
		prop.put("<xmlattr>.type", value_type);
	}

	void operator()(const configuration::string_set& value) const
	{
		boost::property_tree::ptree& prop = settings_.add_child("property", boost::property_tree::ptree());
		for(configuration::string_set::const_iterator it = value.begin(); it != value.end(); ++it)
			prop.add("elem", (*it));

		put_type(prop, "set");
		put_name(prop);
	}

	void operator()(uint32_t value) const
	{
		boost::property_tree::ptree& prop = settings_.add("property", value);
		put_type(prop, "uint32_t");
		put_name(prop);
	}

	void operator()(const std::string& value) const
	{
		boost::property_tree::ptree& prop = settings_.add("property", value);
		put_type(prop, "string");
		put_name(prop);
	}

	void operator()(bool value) const
	{
		boost::property_tree::ptree& prop = settings_.add("property", value);
		put_type(prop, "bool");
		put_name(prop);
	}

private:
	mutable boost::property_tree::ptree& settings_;
	std::string name_;
};

void configuration::save(const std::wstring& file_name, const std::wstring& app_name)
{
	std::ofstream stream(prepare_settings_directory(file_name, app_name, true).c_str(), std::ios::out | std::ios::trunc);
	if(!stream.is_open() || stream.bad())
		throw sp_exception(sp_exception::internal_error, L"Unable to create xml file");

	boost::property_tree::ptree pt;
	boost::property_tree::ptree& settings = pt.put_child("settings", boost::property_tree::ptree());

	for(prop_map::const_iterator it = props_.begin(); it != props_.end(); ++it)
	{
		if(!(*it).second.save_to_file())
			continue;

		boost::apply_visitor(save_visitor(settings, (*it).first), (*it).second.get_variant());
	}

	boost::property_tree::xml_parser::write_xml(stream, pt);
}

const std::wstring configuration::get_local_appdata_path()
{
	wchar_t buf[MAX_PATH] = {};
	if(FAILED(SHGetFolderPathW(0, CSIDL_LOCAL_APPDATA, NULL, SHGFP_TYPE_CURRENT, buf)))
		throw sp_exception(sp_exception::internal_error, L"Unable to get local appdata path");

	return buf;
}

const std::wstring configuration::prepare_settings_directory(const std::wstring& file_name, const std::wstring& app_name,
	bool create)
{
	boost::filesystem::path full_path(get_local_appdata_path());
	full_path /= app_name;
	if(create && !boost::filesystem::is_directory(full_path))
		boost::filesystem::create_directory(full_path);

	full_path /= file_name;

	return full_path.c_str();
}

void configuration::set_loader_and_displayer(web_content_loader& loader, image_window& displayer)
{
	loader_ = &loader;
	displayer_ = &displayer;
}

void configuration::apply_settings()
{
	if(!loader_ || !displayer_)
		return;

	loader_->set_connection_timeout(configuration::get_property("connection_timeout", skype::default_connection_timeout));
	loader_->set_timeout(configuration::get_property("read_timeout", skype::default_read_timeout));
	loader_->set_max_data_size(configuration::get_property("max_data_size", skype::default_max_data_size));
	loader_->set_max_redirects(configuration::get_property("max_redirects", skype::default_max_redirects));
	loader_->set_user_agent(configuration::get_property("user_agent", skype::default_user_agent));

	url_processor::set_max_sizes(configuration::get_property("max_image_x_y_size", skype::default_max_image_x_y_size),
		configuration::get_property("max_gif_x_y_size", skype::default_max_gif_x_y_size));

	url_processor::set_max_title_length(configuration::get_property("max_title_length", skype::default_max_title_length));
	url_processor::set_default_encoding(configuration::get_property("default_encoding", skype::default_encoding));
	url_processor::set_auto_proxy(configuration::get_property("auto_proxy", true));
	url_processor::set_no_proxy(configuration::get_property("no_proxy", false));
	url_processor::set_proxy(configuration::get_property("proxy", std::string()));
	url_processor::set_proxy_auth(configuration::get_property("proxy_auth", false),
		configuration::get_property("proxy_login", std::string()),
		configuration::get_property("proxy_password", std::string()));

	url_processor::set_contains_filter(configuration::get_property("filter_preview_contains", false),
		configuration::get_property("url_contains_set", configuration::string_set()));
	url_processor::set_does_not_contain_filter(configuration::get_property("filter_preview_does_not_contain", false),
		configuration::get_property("url_does_not_contain_set", configuration::string_set()));

	url_processor::enable_preview(configuration::get_property("preview_images", true),
		configuration::get_property("preview_titles", true));

	displayer_->set_min_frame_time(configuration::get_property("min_frame_time", skype::default_min_frame_time));
}
