#include <Windows.h>
#include <GdiPlus.h>
#include "image_window.h"
#include "resource.h"
#include "sp_exception.h"
#include "utility.h"
#include "configuration.h"

namespace
{
const uint32_t timer_event_id = 1000;
const uint32_t message_replace_image = WM_USER + 100;
const uint32_t message_quit = WM_USER + 101;
const uint32_t min_frame_time = 80;
const int32_t delta_x = 10;
const int32_t delta_y = 10;
}

image_window::image_window(unpacked_image* image)
	:image_(image),
	current_frame_(0),
	dialog_(0),
	min_frame_time_(min_frame_time)
{
}

void image_window::set_min_frame_time(uint32_t min_frame_time_value)
{
	min_frame_time_ = min_frame_time_value;
}

void image_window::reset(unpacked_image* image)
{
	SendMessageW(dialog_, message_replace_image, reinterpret_cast<WPARAM>(image), 0);
}

void image_window::show(int32_t x, int32_t y)
{
	const RECT monitor_rect(utility::get_monitor_rect_from_point(x, y));
	
	uint32_t image_width = image_->get_image(0).GetWidth();
	uint32_t image_height = image_->get_image(0).GetHeight();

	if(y + static_cast<int32_t>(image_height) + delta_y > monitor_rect.bottom)
		y = y - image_height - delta_y;
	else
		y += delta_y;

	if(x + static_cast<int32_t>(image_width) + delta_x > monitor_rect.right)
		x = x - image_width - delta_x;
	else
		x += delta_x;

	SetWindowPos(dialog_, HWND_TOPMOST, x, y, 0, 0, SWP_NOSIZE | SWP_NOACTIVATE);
	ShowWindow(dialog_, SW_SHOWNOACTIVATE);
}

void image_window::hide()
{
	ShowWindow(dialog_, SW_HIDE);
}

void image_window::message_queue()
{
	MSG msg;
	while(GetMessage(&msg, NULL, 0, 0) > 0)
	{
		if(!IsDialogMessage(dialog_, &msg))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
}

void image_window::start()
{
	dialog_ = CreateDialogParamW(configuration::get_instance(), MAKEINTRESOURCEW(IDD_PREVIEW),
		NULL, general_preview_func, reinterpret_cast<LPARAM>(this));

	setup_image();

	if(!dialog_)
		throw sp_exception(sp_exception::internal_error, L"Unable to create dialog");
}

void image_window::close()
{
	if(image_->get_frame_count() > 1)
		KillTimer(dialog_, timer_event_id);

	SendMessageW(dialog_, message_quit, 0, 0);
}

void image_window::setup_image()
{
	KillTimer(dialog_, timer_event_id);

	current_frame_ = 0;

	SetWindowPos(dialog_, HWND_TOPMOST, 0, 0,
		image_->get_image(0).GetWidth(), image_->get_image(0).GetHeight(), SWP_NOMOVE | SWP_NOACTIVATE);

	if(image_->get_frame_count() > 1)
	{
		image_->set_min_frame_time(min_frame_time_);
		SetTimer(dialog_, timer_event_id, image_->get_frame_time(0), NULL);
		current_frame_ = 1;
	}

	RedrawWindow(dialog_, NULL, NULL, RDW_INVALIDATE);
	UpdateWindow(dialog_);
}

INT_PTR CALLBACK image_window::preview_func(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM /*lParam*/)
{
	switch(uMsg)
	{
	case WM_MOUSEMOVE:
		ShowWindow(hwndDlg, SW_HIDE);
		break;

	case message_replace_image:
		{
			image_.reset(reinterpret_cast<unpacked_image*>(wParam));
			setup_image();
		}
		break;

	case WM_PAINT:
		{
			uint32_t width = image_->get_image(0).GetWidth();
			uint32_t height = image_->get_image(0).GetHeight();

			Gdiplus::Bitmap buf(width, height);
			
			Gdiplus::Graphics graphics(&buf);
			Gdiplus::Rect dest(0, 0, width, height);

			Gdiplus::SolidBrush brush(Gdiplus::Color(255, 255, 255));
			graphics.FillRectangle(&brush, dest);

			graphics.DrawImage(&image_->get_image(current_frame_), dest, 0, 0, width, height, Gdiplus::UnitPixel);
			
			PAINTSTRUCT ps;
			HDC dc = BeginPaint(hwndDlg, &ps);
			
			Gdiplus::Graphics hdc_graphics(dc);
			hdc_graphics.DrawImage(&buf, dest, 0, 0, width, height, Gdiplus::UnitPixel);

			EndPaint(hwndDlg, &ps);
		}
		break;

	case message_quit:
		DestroyWindow(hwndDlg);
		break;

	case WM_DESTROY:
		PostQuitMessage(0);
		break;

	case WM_TIMER:
		{
			uint32_t frame_count = image_->get_frame_count();
			if(frame_count > 1)
			{
				current_frame_ = (current_frame_ + 1) % frame_count;
				SetTimer(hwndDlg, timer_event_id, image_->get_frame_time(current_frame_), NULL);
				RedrawWindow(hwndDlg, NULL, NULL, RDW_INVALIDATE);
				UpdateWindow(hwndDlg);
			}
		}
		break;

	default:
		return FALSE;
	}

	return TRUE;
}

INT_PTR CALLBACK image_window::general_preview_func(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
	case WM_INITDIALOG:
		SetWindowLongPtrW(hwndDlg, GWLP_USERDATA, lParam);
		break;

	default:
		{
			image_window* win = reinterpret_cast<image_window*>(GetWindowLongPtrW(hwndDlg, GWLP_USERDATA));
			return win ? win->preview_func(hwndDlg, uMsg, wParam, lParam) : FALSE;
		}
	}

	return TRUE;
}
