#include <locale>
#include <codecvt>
#include <boost/bind.hpp>
#include "url_processor.h"
#include "web_content_loader.h"
#include "image_window.h"
#include "utility.h"
#include "istream_owner_image.h"
#include "byte_stream.h"
#include "sp_exception.h"

boost::condition_variable url_processor::worker_var_;
boost::mutex url_processor::worker_mutex_;
std::wstring url_processor::last_url_;
std::wstring url_processor::prev_url_;
uint32_t url_processor::max_size_ = 350;
uint32_t url_processor::max_gif_size_ = 650;
uint32_t url_processor::max_title_length_ = 100;
boost::thread url_processor::processor_thread_;
web_content_loader* url_processor::loader_;
image_window* url_processor::displayer_;
int32_t url_processor::last_x_ = 0;
int32_t url_processor::last_y_ = 0;
bool url_processor::active_cursor_ = false;
std::auto_ptr<unpacked_image> url_processor::wait_image_;
std::string url_processor::default_encoding_;
std::string url_processor::proxy_;
bool url_processor::auto_proxy_ = true;
bool url_processor::no_proxy_ = false;
bool url_processor::use_proxy_auth_ = false;
std::string url_processor::proxy_login_;
std::string url_processor::proxy_password_;
bool url_processor::preview_images_ = true;
bool url_processor::preview_titles_ = true;
bool url_processor::contains_filter_on_ = false;
bool url_processor::does_not_contain_filter_on_ = false;
configuration::string_set url_processor::contains_filter_strings_;
configuration::string_set url_processor::does_not_contain_filter_strings_;

void url_processor::set_contains_filter(bool filter_on, const configuration::string_set& strings)
{
	contains_filter_on_ = filter_on;
	contains_filter_strings_ = strings;
}

void url_processor::set_does_not_contain_filter(bool filter_on, const configuration::string_set& strings)
{
	does_not_contain_filter_on_ = filter_on;
	does_not_contain_filter_strings_ = strings;
}

void url_processor::enable_preview(bool preview_images, bool preview_titles)
{
	preview_images_ = preview_images;
	preview_titles_ = preview_titles;
}

void url_processor::set_proxy_auth(bool use_auth, const std::string& login, const std::string& password)
{
	use_proxy_auth_ = use_auth;
	proxy_login_ = login;
	proxy_password_ = password;
}

void url_processor::set_proxy(const std::string& proxy)
{
	proxy_ = proxy;
}

void url_processor::set_auto_proxy(bool auto_proxy)
{
	auto_proxy_ = auto_proxy;
}

void url_processor::set_no_proxy(bool no_proxy)
{
	no_proxy_ = no_proxy;
}

void url_processor::set_default_encoding(const std::string& default_encoding)
{
	default_encoding_ = default_encoding;
}

void url_processor::set_max_title_length(uint32_t max_title_length)
{
	max_title_length_ = max_title_length;
}

void url_processor::set_max_sizes(uint32_t max_size, uint32_t max_gif_size)
{
	max_size_ = max_size;
	max_gif_size_ = max_gif_size;
}

void url_processor::process_url(web_content_loader& loader,
	image_window& displayer, const std::wstring& url, int32_t x, int32_t y)
{
	std::string url_utf8;

	{
		std::wstring_convert<std::codecvt_utf8_utf16<wchar_t> > converter;

		if(auto_proxy_)
		{
			std::wstring proxy, bypass;
			utility::get_proxy_info(proxy, bypass);
			std::string proxy_utf8 = converter.to_bytes(proxy);
			loader.set_proxy(proxy_utf8);
		}
		else if(no_proxy_)
		{
			loader.set_proxy(std::string());
		}
		else
		{
			loader.set_proxy(proxy_);

			if(use_proxy_auth_)
				loader.set_proxy_auth(proxy_login_, proxy_password_);
			else
				loader.set_proxy_auth(std::string(), std::string());
		}

		url_utf8 = converter.to_bytes(url);
	}
	
	std::string headers;
	std::string res(loader.load(url_utf8, &headers));

	bool show_displayer = false;

	{
		std::auto_ptr<istream_owner_image> img(new istream_owner_image(byte_stream::create(res), false));
		if(img->GetLastStatus() == Gdiplus::Ok)
		{
			if(preview_images_)
			{
				std::auto_ptr<unpacked_image> img(utility::scale_image(*img, max_size_, max_gif_size_));
				displayer.hide();
				displayer.reset(img.release());
				show_displayer = true;
			}
		}
		else if(preview_titles_)
		{
			const std::string title(utility::get_html_page_title(res));
			if(title.empty())
				throw sp_exception(sp_exception::content_not_supported, L"Not supported content");

			std::string charset(utility::get_html_page_charset(headers, res));
			if(charset.empty())
			{
				if(default_encoding_.empty())
					throw sp_exception(sp_exception::content_not_supported, L"Unknown content charset");

				charset = default_encoding_;
			}

			std::wstring title_ucs2(utility::trim(
				utility::replace_html_entities(
				utility::convert_to_unicode(title, charset))));
			if(title_ucs2.length() > max_title_length_)
				title_ucs2 = title_ucs2.substr(0, max_title_length_) + L"...";

			show_displayer = true;
			displayer.hide();
			displayer.reset(new unpacked_image(utility::draw_text(title_ucs2, max_size_).release()));
		}
	}

	if(show_displayer)
		displayer.show(x, y);
	else
		displayer.hide();
}

void url_processor::set_wait_image(Gdiplus::Image* image)
{
	if(image)
	{
		uint32_t max_size = (std::max)(image->GetWidth(), image->GetHeight());
		wait_image_.reset(utility::scale_image(*image, max_size, max_size).release());
	}
	else
	{
		wait_image_.reset(0);
	}
}

void url_processor::set_work_item(const std::wstring& url, int32_t x, int32_t y, bool active_cursor)
{
	if(contains_filter_on_ || does_not_contain_filter_on_)
	{
		std::wstring_convert<std::codecvt_utf8_utf16<wchar_t> > converter;
		std::string url_utf8(converter.to_bytes(url));

		bool ok_containt = true;
		bool ok_does_not_contain = true;

		if(contains_filter_on_)
		{
			ok_containt = false;

			for(configuration::string_set::const_iterator it = contains_filter_strings_.begin();
				it != contains_filter_strings_.end(); ++it)
			{
				if(url_utf8.find(*it) != std::string::npos)
				{
					ok_containt = true;
					break;
				}
			}
		}

		if(does_not_contain_filter_on_)
		{
			ok_does_not_contain = true;

			for(configuration::string_set::const_iterator it = does_not_contain_filter_strings_.begin();
				it != does_not_contain_filter_strings_.end(); ++it)
			{
				if(url_utf8.find(*it) != std::string::npos)
				{
					ok_does_not_contain = false;
					break;
				}
			}
		}

		if(!ok_containt || !ok_does_not_contain)
			return;
	}

	{
		boost::lock_guard<boost::mutex> lock(worker_mutex_);
		last_url_ = url;
		last_x_ = x;
		last_y_ = y;
		active_cursor_ = active_cursor;
		displayer_->hide();

		if(!loader_->is_ready())
			loader_->abort();

		if(active_cursor && wait_image_.get())
		{
			displayer_->reset(wait_image_->clone().release());
			displayer_->show(x, y);
		}
	}

	worker_var_.notify_one();
}

void url_processor::monitor_thread(web_content_loader& loader, image_window& displayer)
{
	while(true)
	{
		boost::unique_lock<boost::mutex> lock(worker_mutex_);
		while(last_url_.empty())
			worker_var_.wait(lock);

		std::wstring url = last_url_;
		int32_t x = last_x_;
		int32_t y = last_y_;
		bool active_cursor = active_cursor_;
		last_url_.clear();

		lock.unlock();
		
		if(!active_cursor)
		{
			if(url == prev_url_)
				displayer.hide();

			continue;
		}
		

		prev_url_ = url;

		try
		{
			process_url(loader, displayer, url, x, y);
		}
		catch(const std::exception&)
		{
			displayer.hide();
		}
	}
}

void url_processor::start(web_content_loader& loader, image_window& displayer)
{
	loader_ = &loader;
	displayer_ = &displayer;
	processor_thread_ = boost::thread(boost::bind(monitor_thread, boost::ref(loader), boost::ref(displayer)));
}

void url_processor::wait()
{
	if(processor_thread_.joinable())
		processor_thread_.join();
}
