#include <curl/curl.h>
#include <boost/thread/mutex.hpp>
#include <boost/algorithm/string.hpp>
#include "web_content_loader.h"
#include "sp_exception.h"

uint32_t web_content_loader::global_init_ = 0;
boost::mutex web_content_loader::global_init_mutex_;

web_content_loader::web_content_loader()
	:curl_(0),
	aborted_(false),
	ready_(true),
	max_data_size_(static_cast<uint32_t>(-1))
{
	{
		boost::mutex::scoped_lock lock(global_init_mutex_);
		if(!global_init_)
		{
			curl_global_init(CURL_GLOBAL_ALL);
			++global_init_;
		}
	}

	curl_ = curl_easy_init();

	if(!curl_)
	{
		cleanup();
		throw sp_exception(sp_exception::internal_error, L"Unable to init CURL");
	}

	curl_easy_setopt(curl_, CURLOPT_FOLLOWLOCATION, 1);
	curl_easy_setopt(curl_, CURLOPT_REDIR_PROTOCOLS, CURLPROTO_HTTP | CURLPROTO_HTTPS);
	curl_easy_setopt(curl_, CURLOPT_COOKIEFILE, "");
	curl_easy_setopt(curl_, CURLOPT_NOPROGRESS, 0);
	curl_easy_setopt(curl_, CURLOPT_PROGRESSDATA, this);
	curl_easy_setopt(curl_, CURLOPT_PROGRESSFUNCTION, progress_callback);
	curl_easy_setopt(curl_, CURLOPT_WRITEFUNCTION, data_callback);
	curl_easy_setopt(curl_, CURLOPT_SSL_VERIFYPEER, 0);
}

void web_content_loader::set_user_agent(const std::string& user_agent)
{
	curl_easy_setopt(curl_, CURLOPT_USERAGENT, user_agent.c_str());
}

void web_content_loader::set_max_redirects(uint32_t max_redirects)
{
	curl_easy_setopt(curl_, CURLOPT_MAXREDIRS, max_redirects);
}

void web_content_loader::set_max_data_size(uint32_t max_data_size)
{
	max_data_size_ = max_data_size;
}

void web_content_loader::abort()
{
	boost::mutex::scoped_lock lock(abort_mutex_);
	aborted_ = true;
}

bool web_content_loader::is_ready() const
{
	boost::mutex::scoped_lock lock(ready_mutex_);
	return ready_;
}

int web_content_loader::progress_callback(void *clientp, double /*dltotal*/, double dlnow, double /*ultotal*/, double /*ulnow*/)
{
	web_content_loader* loader = reinterpret_cast<web_content_loader*>(clientp);

	if(dlnow > loader->max_data_size_)
		return 1;

	{
		boost::mutex::scoped_lock lock(loader->abort_mutex_);
		return loader->aborted_ ? 1 : 0;
	}
}

void web_content_loader::cleanup()
{
	if(curl_)
	{
		curl_easy_cleanup(curl_);
		curl_ = 0;
	}

	{
		boost::mutex::scoped_lock lock(global_init_mutex_);
		if(!(--global_init_))
			curl_global_cleanup();
	}
}

web_content_loader::~web_content_loader()
{
	cleanup();
}

size_t web_content_loader::data_callback(void *ptr, size_t size, size_t nmemb, void *stream)
{
	std::string* data = reinterpret_cast<std::string*>(stream);
	if(data)
		data->append(reinterpret_cast<const char*>(ptr), size * nmemb);

	return size * nmemb;
}

const std::string web_content_loader::load(const std::string& url, std::string* headers)
{
	{
		boost::mutex::scoped_lock lock(ready_mutex_);
		if(!ready_)
			throw sp_exception(sp_exception::unable_to_load_content_not_ready, L"Unable to load content with CURL - not ready");

		ready_ = false;
	}

	{
		boost::mutex::scoped_lock lock(abort_mutex_);
		aborted_ = false;
	}

	std::string data;
	curl_easy_setopt(curl_, CURLOPT_URL, url.c_str());
	curl_easy_setopt(curl_, CURLOPT_COOKIELIST, "ALL");
	curl_easy_setopt(curl_, CURLOPT_WRITEDATA, &data);

	if(headers)
	{
		headers->clear();
		curl_easy_setopt(curl_, CURLOPT_WRITEHEADER, headers);
	}
	else
	{
		curl_easy_setopt(curl_, CURLOPT_WRITEHEADER, 0);
	}

	CURLcode res = curl_easy_perform(curl_);
	
	{
		boost::mutex::scoped_lock lock(ready_mutex_);
		ready_ = true;
	}

	if(res != CURLE_OK)
		throw sp_exception(sp_exception::unable_to_load_content, L"Unable to load content with CURL");

	return data;
}

void web_content_loader::set_timeout(uint32_t timeout)
{
	curl_easy_setopt(curl_, CURLOPT_TIMEOUT, timeout);
}

void web_content_loader::set_connection_timeout(uint32_t connection_timeout)
{
	curl_easy_setopt(curl_, CURLOPT_CONNECTTIMEOUT, connection_timeout);
}

void web_content_loader::set_proxy(const std::string& proxy_url)
{
	curl_easy_setopt(curl_, CURLOPT_PROXY, proxy_url.c_str());
}

void web_content_loader::set_proxy_auth(const std::string& proxy_login, const std::string& proxy_password)
{
	if(proxy_login.empty() && proxy_password.empty())
	{
		curl_easy_setopt(curl_, CURLOPT_PROXYUSERPWD, "");
		curl_easy_setopt(curl_, CURLOPT_PROXYAUTH, 0);
	}
	else
	{
		curl_easy_setopt(curl_, CURLOPT_PROXYUSERPWD, (proxy_login + ":" + proxy_password).c_str());
		curl_easy_setopt(curl_, CURLOPT_PROXYAUTH, CURLAUTH_ANY);
	}
}

const web_content_loader::proxy_info web_content_loader::get_proxy_info(const std::wstring& proxy_string)
{
	proxy_type type = proxy_http;
	uint32_t skip = 0;

	if(boost::istarts_with(proxy_string, L"socks4://"))
	{
		skip = sizeof("socks4://") - 1;
		type = proxy_socks4;
	}
	else if(boost::istarts_with(proxy_string, L"socks4a://"))
	{
		skip = sizeof("socks4a://") - 1;
		type = proxy_socks4a;
	}
	else if(boost::istarts_with(proxy_string, L"socks5://"))
	{
		skip = sizeof("socks5://") - 1;
		type = proxy_socks5;
	}
	else if(boost::istarts_with(proxy_string, L"socks5h://"))
	{
		skip = sizeof("socks5h://") - 1;
		type = proxy_socks5h;
	}

	return std::make_pair(std::wstring(proxy_string.c_str() + skip), type);
}
